import logging.config


def configure_logging(log_level, log_format, log_custom_format=None):
    log_custom_format = log_custom_format or log_format
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'quiet': {
                'format': '%(asctime)s [%(levelname)s] %(message)s'
            },
            'verbose': {
                'format': 'PID%(process)d %(asctime)s [%(levelname)s:%(name)s:%(lineno)d] %(message)s'
            },
            'custom': {
                'format': log_custom_format
            }
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': log_format
            }
        },
        'loggers': {
            'logdog': {
                'handlers': ['console'],
                'level': log_level,
                'propagate': False,
            },
            'tornado': {
                'handlers': ['console'],
                'level': log_level,
                'propagate': False,
            },
            'zmq': {
                'handlers': ['console'],
                'level': log_level,
                'propagate': False,
            },
        }
    })




"""logdog command line interface

Usage:
  logdog [<pipe-namespace>...] [options]
  logdog (-h | --help)
  logdog --version

Arguments:
  <pipe-namespace>          One or more pipe namespaces to be run

Options:
  -h --help                 Show this screen
  --version                 Show version
  -v --verbose              Run in verbose mode
  -l --log-level=<level>    Set logging level [default: INFO]
  -f --log-format=<format>  Set logging format [default: quiet]
  -c --config=<config>      Configuration file (yaml config)
  -s --sources=<file:...>   Force specify files to be watched
  -H --handler=<handler>    Force set handler for all sources
                            (e.g. --handler=viewers.console)
  --reset-indices           Remove current indices.
"""
from docopt import docopt
from logdog.app import Application
from logdog.core.config import ConfigLoader
from logdog.core.log import configure_logging
from logdog.version import __version__


def main():
    arguments = docopt(__doc__, version='logdog {}'.format(__version__))

    log_level = arguments.get('--log-level').upper()
    log_format = log_custom_format = arguments.get('--log-format')

    if log_format not in ('verbose', 'quiet'):
        log_format = 'custom'

    if arguments.get('--verbose'):
        log_level = 'DEBUG'
        log_format = 'verbose'

    configure_logging(log_level, log_format, log_custom_format)

    config_path = arguments.get('--config')
    loader = ConfigLoader(path=config_path)
    config = loader.load_config(default_only=not config_path)

    Application(
        active_namespaces=arguments.get('<pipe-namespace>'),
        config=config,
        force_handler=arguments.get('--handler'),
        force_sources=arguments.get('--sources'),
        reset_indices=arguments.get('--reset-indices'),
    ).run()

if __name__ == '__main__':
    main()


import logging
from string import ascii_letters
import time
import functools


logger = logging.getLogger(__name__)


def mark_as_coroutine(func):
    func.is_coroutine = True
    return func


def mark_as_proxy_method(func):
    func.is_proxy_method = True
    return func


def is_coroutine(func):
    return getattr(func, 'is_coroutine', False)


def is_proxy(func):
    return getattr(func, 'is_proxy_method', False)


def simple_oid():
    oid = []
    x = int(time.time() * 10000) % 10 ** 8
    n = len(ascii_letters)
    while x:
        oid.insert(0, ascii_letters[x % n])
        x /= n
    return ''.join(oid)


def debug_deco(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.exception(e)
    return wrapper


import json
import umsgpack


class Msg(object):
    __slots__ = (
        'message',
        'source',
        'meta',
        'version',
    )

    def __init__(self, message, source, meta=None, version=1):
        self.message = message
        self.source = source
        self.meta = meta
        self.version = version

    def __str__(self):
        return self.message

    def serialize(self):
        return {
            'msg': self.message,
            'src': self.source,
            'meta': self.meta,
            '_v': self.version,
        }

    @classmethod
    def deserialize(cls, data):
        return cls(message=data.get('msg'),
                   source=data.get('src'),
                   meta=data.get('meta'),
                   version=data.get('_v'))

    def serialize_json(self):
        return json.dumps(self.serialize())

    @classmethod
    def deserialize_json(cls, data):
        return cls.deserialize(json.loads(data))

    def serialize_jsonb(self):
        return umsgpack.dumps(self.serialize())

    @classmethod
    def deserialize_jsonb(cls, data):
        return cls.deserialize(umsgpack.loads(data))



# options must be valid python identifiers, e.g. "default_pipe"
# names can be any string, e.g. "file->webui" (excluding ' ', '@')

config = {
    'sources': {
        '/var/log/*.log': {'handler': 'pipes.to-web'},
        '/var/log/*/*.log': {'handler': 'pipes.to-web'},
        '/var/log/syslog': 'pipes.to-web',
    },

    'pipes': {
        'default': 'logdog.pipes.Pipe',

        'to-web': [
            'watch processors.stripper',
            'watch connectors.zmq-tunnel@sender',
            'view connectors.zmq-tunnel@receiver',
            'view viewers.webui',
        ],
        'experiment-x001': {
            'cls': 'logdog.pipes.Pipe',
            '*': [
                'watch processors.stripper',
                {'forwarders.broadcast': [
                    'watch viewers.console',
                    {'pipes.default': [
                        {'watch connectors.zmq-tunnel@sender': {'connect': 'tcp://localhost:7789'}},
                        {'view connectors.zmq-tunnel@receiver': {'bind': 'tcp://*:7789'}},
                        {'view forwarders.round-robin': [
                            'view viewers.webui',
                            'view viewers.null',
                        ]}
                    ]},
                ]},
            ],
        }
    },

    'options': {
        'sources': {
            'default_handler': 'pipes.to-web',
            'default_watcher': 'pollers.file-watcher',
            'index_file': '~/.config/logdog/sources-index.idx'
        }
    },

    # pollers
    'pollers': {
        'file-watcher': {
            'cls': 'logdog.pollers.FileWatcher',
            'namespaces': ['watch'],
        },
    },

    # collectors
    'collectors': {},

    # processors
    'processors': {
        'stripper': {'cls': 'logdog.processors.Stripper'},
    },

    # forwarders
    'forwarders': {
        'broadcast': 'logdog.forwarders.Broadcast',
        'round-robin': 'logdog.forwarders.RoundRobin',
    },

    # connectors
    'connectors': {
        'zmq-tunnel': {
            'cls': 'logdog.connectors.ZMQTunnel',
            '@sender': {'socket': 'PUSH', 'connect': ['tcp://localhost:45457']},
            '@receiver': {'socket': 'PULL', 'bind': ['tcp://*:45457']},
        }
    },

    # viewers
    'viewers': {
        'default': 'logdog.viewers.Null',

        'webui': {
            'cls': 'logdog.viewers.WebUI',
            'port': 8888,
        },
        'console': {
            'cls': 'logdog.viewers.Console',
            'redirect_to': 'stdout',
        },
        'null': 'logdog.viewers.Null',
    },

    # utils
    'utils': {
        'policies': {
            'growing-sleep': {
                'cls': 'logdog.core.policies.DefaultSleepPolicy'
            },
            'mostly-greedy': {
                'cls': 'logdog.core.policies.GreedyPolicy'
            }
        }
    }
}



from logdog.core.roles.processor import BaseProcessor


class Stripper(BaseProcessor):

    def __str__(self):
        return u'STRIPPER'

    def on_recv(self, data):
        data.message = data.message.strip()
        return self.output.send(data)


import logging
import os
import sys

from logdog.core.roles.viewer import BaseViewer


logger = logging.getLogger(__name__)


class Console(BaseViewer):
    defaults = BaseViewer.defaults.copy_and_update(
        redirect_to='stdout'
    )

    def __init__(self, *args, **kwargs):
        super(Console, self).__init__(*args, **kwargs)
        self._output = getattr(sys, self.config.redirect_to)

    def _input_forwarder(self, data):
        sys.stdout.write(data.message)
        if not data.message.endswith(os.linesep):
            sys.stdout.write(os.linesep)

    on_recv = _input_forwarder


import logging

from tornado import gen
from tornado.web import RequestHandler
from tornado.websocket import WebSocketHandler


logger = logging.getLogger(__name__)


class IndexHandler(RequestHandler):

    def get(self):
        self.render('index.html')


class LogWebSocket(WebSocketHandler):

    @gen.coroutine
    def open(self):
        logger.info('Client connected.')
        self.application.ws.append(self)

    def on_close(self):
        logger.debug('Connection closed.')
        self.application.ws.remove(self)


MAJOR_VERSION = 0
MINOR_VERSION = 2
PATCH_VERSION = 4
version = (MAJOR_VERSION, MINOR_VERSION, PATCH_VERSION)

__version__ = '.'.join(map(str, version))
__author__ = 'Evgeny Lychkovsky'
__email__ = 'miphreal@gamil.com'



import logging
from tornado import gen
from tornado.concurrent import is_future

from logdog.core.roles.pollers import BasePoller


logger = logging.getLogger(__name__)


class FileWatcher(BasePoller):
    defaults = BasePoller.defaults.copy_and_update(
        poll_sleep_policy='utils.policies.growing-sleep',
        greedy_policy='utils.policies.mostly-greedy',
        greedy_file_reading=True,
        start_delay=2.5,
    )

    def __init__(self, *args, **kwargs):
        super(FileWatcher, self).__init__(*args, **kwargs)
        self.poll_sleep_policy = self.app.config.find_and_construct_class(name=self.config.poll_sleep_policy)
        self.greedy_policy = self.app.config.find_and_construct_class(name=self.config.greedy_policy)

    def __str__(self):
        return u'WATCHER:{!s}'.format(self.input)

    def _prepare_message(self, data):
        msg = super(FileWatcher, self)._prepare_message(data)
        msg.source = getattr(self.input, 'name', msg.source)
        return msg

    @gen.coroutine
    def poll(self):
        greedy_file_reading_enabled = self.config.greedy_file_reading

        try:
            self.input.open()
            logger.info('[%s] Opened %s.', self, self.input)
        except IOError as e:
            logger.error('[%s] Stopping watching because of errors (%s)...', self, e)
            self.started = False
            yield self.stop()
            return

        while self.started:
            data = self.input.read_line()

            if data:
                data = self._prepare_message(data)
                ret = self._forward(data)
                if is_future(ret):
                    yield ret

                self.poll_sleep_policy.reset()

                if greedy_file_reading_enabled and self.greedy_policy.need_to_wait():
                    yield self.greedy_policy.wait()

            else:
                self.input.check_stat()
                self.app.register.set_path(self.input)
                logger.debug('[%s] Sleep on watching %ss.', self, self.poll_sleep_policy.cur_interval)
                yield self.poll_sleep_policy.sleep()


import logging
from tornado import gen

from ..utils import mark_as_proxy_method
from .base import BaseRole


logger = logging.getLogger(__name__)


class BasePoller(BaseRole):
    def poll(self):
        # Straightforward implementation
        while self.started:
            self.send(self.input.pop())

    @mark_as_proxy_method
    def on_recv(self, data):
        return self.output.send(data)

    @gen.coroutine
    def start(self):
        yield super(BasePoller, self).start()
        if self.started:
            yield self.poll()



import logging
import os


logger = logging.getLogger(__name__)


class Path(object):
    __slots__ = ('name', 'stat', '_prev_stat', 'offset', '_f', '_last_read_line')

    def __init__(self, name, offset, stat):
        self.name = name
        self.stat = self._prev_stat = stat
        self.offset = offset

        self._f = self._last_read_line = None

    def __str__(self):
        return self.name

    def __getstate__(self):
        return {
            'name': self.name,
            'stat': self.stat,
            '_prev_stat': self._prev_stat,
            'offset': self.offset,
            '_last_read_line': self._last_read_line,
        }

    def __setstate__(self, value):
        self.name = value['name']
        self.stat = value['stat']
        self.offset = value['offset']
        self._prev_stat = value['_prev_stat']
        self._last_read_line = value['_last_read_line']
        self._f = None

    def open(self):
        self._f = open(self.name, 'r')
        self._f.seek(self.offset)
        self._prev_stat = self.stat
        self.stat = os.stat(self.name)

    def close(self):
        self._f.close()

    def reopen(self):
        self.close()
        self.open()

    def check_stat(self):
        """Is called if we have nothing to read"""
        self._prev_stat = self.stat
        self.stat = os.stat(self.name)

        if not self.is_same_file(self._prev_stat, self.stat):
            logger.debug('[FILE:%s] New file detected by this path. Re-opening...', self.name)
            self.offset = 0
            self.reopen()

        elif self.is_file_truncated(self._prev_stat, self.stat):
            logger.debug('[FILE:%s] Seems, file was truncated. Re-opening...', self.name)
            self.offset = 0
            self.reopen()

    @staticmethod
    def is_same_file(prev_stat, cur_stat):
        return prev_stat.st_ino == cur_stat.st_ino and prev_stat.st_dev == cur_stat.st_dev

    @staticmethod
    def is_file_truncated(prev_stat, cur_stat):
        return prev_stat.st_size > cur_stat.st_size

    def read_line(self):
        self._last_read_line = (_, line) = (self.offset, self._f.readline())
        self.offset = self._f.tell()
        return line
